@@ -52,20 +52,30 @@ def tensorleap_custom_metric(name: str, direction: Optional[MetricDirection] = M
                                        f'Metric has returned unsupported type. Supported types are List[float], '
                                        f'List[List[ConfusionMatrixElement]], NDArray[np.float32]. ')
 
-            if isinstance(result, list):
-                if isinstance(result[0], list):
-                    assert isinstance(result[0][0], ConfusionMatrixElement), \
-                        f'{supported_types_message}Got List[List[{type(result[0][0])}]].'
+            def _validate_single_metric(single_metric_result):
+                if isinstance(single_metric_result, list):
+                    if isinstance(single_metric_result[0], list):
+                        assert isinstance(single_metric_result[0][0], ConfusionMatrixElement), \
+                            f'{supported_types_message}Got List[List[{type(single_metric_result[0][0])}]].'
+                    else:
+                        assert isinstance(single_metric_result[0], float), f'{supported_types_message}Got List[{type(single_metric_result[0])}].'
                 else:
-                    assert isinstance(result[0], float), f'{supported_types_message}Got List[{type(result[0])}].'
+                    assert isinstance(single_metric_result, np.ndarray), f'{supported_types_message}Got {type(single_metric_result)}.'
+                    assert len(single_metric_result.shape) == 1, (f'tensorleap_custom_metric validation failed: '
+                                                    f'The return shape should be 1D. Got {len(single_metric_result.shape)}D.')
 
+                if leap_binder.batch_size_to_validate:
+                    assert len(single_metric_result) == leap_binder.batch_size_to_validate, \
+                        f'tensorleap_custom_metrix validation failed: The return len should be as the batch size.'
+
+            if isinstance(result, dict):
+                for key, value in result.items():
+                    assert isinstance(key, str), \
+                        (f'tensorleap_custom_metric validation failed: '
+                         f'Keys in the return dict should be of type str. Got {type(key)}.')
+                    _validate_single_metric(value)
             else:
-                assert isinstance(result, np.ndarray), f'{supported_types_message}Got {type(result)}.'
-                assert len(result.shape) == 1, (f'tensorleap_custom_metric validation failed: '
-                                                f'The return shape should be 1D. Got {len(result.shape)}D.')
-            if leap_binder.batch_size_to_validate:
-                assert len(result) == leap_binder.batch_size_to_validate, \
-                    f'tensorleap_custom_metrix validation failed: The return len should be as the batch size.'
+                _validate_single_metric(result)
 
         def inner(*args, **kwargs):
             _validate_input_args(*args, **kwargs)
@@ -235,14 +245,16 @@ def tensorleap_unlabeled_preprocess():
     return decorating_function
 
 
-def tensorleap_input_encoder(name: str):
+def tensorleap_input_encoder(name: str, channel_dim=-1):
     def decorating_function(user_function: SectionCallableInterface):
         for input_handler in leap_binder.setup_container.inputs:
             if input_handler.name == name:
                 raise Exception(f'Input with name {name} already exists. '
                                 f'Please choose another')
+        if channel_dim <= 0 and channel_dim != -1:
+            raise Exception(f"Channel dim for input {name} is expected to be either -1 or positive")
 
-        leap_binder.set_input(user_function, name)
+        leap_binder.set_input(user_function, name, channel_dim=channel_dim)
 
         def _validate_input_args(sample_id: Union[int, str], preprocess_response: PreprocessResponse):
             assert isinstance(sample_id, (int, str)), \
@@ -263,6 +275,8 @@ def tensorleap_input_encoder(name: str):
             assert result.dtype == np.float32, \
                 (f'tensorleap_input_encoder validation failed: '
                  f'The return type should be a numpy array of type float32. Got {result.dtype}.')
+            assert channel_dim - 1 <= len(result.shape), (f'tensorleap_input_encoder validation failed: '
+                 f'The channel_dim ({channel_dim}) should be <= to the rank of the resulting input rank ({len(result.shape)}).')
 
         def inner(sample_id, preprocess_response):
             _validate_input_args(sample_id, preprocess_response)
@@ -347,9 +361,9 @@ def tensorleap_custom_loss(name: str):
                 (f'tensorleap_custom_loss validation failed: '
                  f'The return type should be a numpy array or a tensorflow tensor. Got {type(result)}.')
 
-        def inner(sample_id, preprocess_response):
-            _validate_input_args(sample_id, preprocess_response)
-            result = user_function(sample_id, preprocess_response)
+        def inner(*args, **kwargs):
+            _validate_input_args(*args, **kwargs)
+            result = user_function(*args, **kwargs)
             _validate_result(result)
             return result
 
